;;/* 
;;    (C) Vasil S. Diadov 2009 
;;    this module is under GPLv2
;;*/

;;Usage:
;;(lzo-make-temp-buffer) creates u8vector to use in compressor as
;;                       temporary dictionary mem block
;;
;;(lzo-compress-u8vector vector temp-buffer) compresses given vector
;;                                           and returs compressed vector
;;(lzo-decompress-u8vector compressed-vector) decompresses given vector
;;                                            and returns decompressed vector
;;
;; On my Core 2 Duo 2.2 GHz laptop speed of compression is near 77 Mb/s of
;; binary data (list of serialized interpreted procedures) and about 103 Mb/s
;; of decompression (103 Mb/s - speed of compressed stream,
;; uncompressed stream speed is about 220 Mb/s)
;; compression ratio ~2.2

(##include "~~/lib/_gambit#.scm")
(##include "~~/lib/gambit#.scm")

(c-declare #<<c-declare-end

#include "minilzo.c"

c-declare-end
)

(define-prim (lzo-make-temp-buffer)
    (make-u8vector (##c-code "___RESULT = ___FIX((LZO1X_1_MEM_COMPRESS+4));")))

(define-prim (lzo-compress-u8vector vect buffer)
    (##declare (not interrupts-enabled))
    (let ((result 
            (make-u8vector 
                (+ (u8vector-length vect) 
                    (arithmetic-shift (u8vector-length vect) -4) 
                    64 3 5)))) ;;extra 5 bytes are for flag of compression and 
                               ;;input length if needed
        (let ((output 
    (##c-code #<<end-of-c-code

            ___SCMOBJ input_vector = ___ARG1;
            ___SCMOBJ work_buffer = ___ARG2;
            ___SCMOBJ output_vector = ___ARG3;
            ___SCMOBJ ___temp;
            lzo_bytep input;
            lzo_bytep output;
            int input_size = ___INT ( ___U8VECTORLENGTH(input_vector));
            int output_size = ___INT ( ___U8VECTORLENGTH(output_vector));
            lzo_uint result_size;
            lzo_voidp wrkmem = (lzo_voidp)___BODY_AS(work_buffer,___tSUBTYPED);
            input = (lzo_bytep) ___BODY_AS(input_vector,___tSUBTYPED);
            output =(lzo_bytep) ___BODY_AS(output_vector,___tSUBTYPED);  
            if ( lzo1x_1_compress(
                    input,
                    (lzo_uint) input_size,
                    output + 5,
                    &result_size,
                    wrkmem) != LZO_E_OK){
                ___U8VECTORSHRINK(output_vector,___FIX_0);
                ___RESULT = ___FIX(___CTOS_HEAP_OVERFLOW_ERR);
            } else {
                if ( result_size >= (lzo_uint) input_size) {
                    int i;
                    *output++ = 0; //uncompressed
                    for(i=0; i<input_size; i++)
                        *output++=*input++;
                    result_size = input_size + 1;
                } else {
                    output[0]=1; //compressed
                    output[1]=input_size&0xFF;
                    output[2]=(input_size>>8)&0xFF;
                    output[3]=(input_size>>16)&0xFF;
                    output[4]=(input_size>>24)&0xFF;
                    result_size+=5;
                }
                ___U8VECTORSHRINK(output_vector,___FIX(result_size));
                ___RESULT = output_vector;
            }
end-of-c-code
    vect buffer result)))
    output)))

(define-prim (lzo-decompress-u8vector vect)
    (##declare (not interrupts-enabled))
    (let ((output-len 
            (##c-code #<<end-of-c-code

                lzo_bytep input = (lzo_bytep) ___BODY_AS(___ARG1,___tSUBTYPED);
                if(0 == *input){
                    //uncompressed
                    ___RESULT = ___FIXSUB(
                                    ___U8VECTORLENGTH(___ARG1),
                                    ___FIX_1);
                }else{
                    input++;
                    ___RESULT = ___FIX (
                                    input[0]+
                                    (input[1]<<8)+
                                    (input[2]<<16)+
                                    (input[3]<<24));
                }

end-of-c-code
vect)))
        (let ((result (make-u8vector output-len)))
            (##c-code #<<end-of-c-code

                lzo_bytep input = (lzo_bytep) ___BODY_AS(___ARG1,___tSUBTYPED);
                lzo_bytep output = (lzo_bytep) ___BODY_AS(___ARG2,___tSUBTYPED);
                int input_size = ___INT ( ___U8VECTORLENGTH( ___ARG1 ));
                lzo_uint output_size;
                ___RESULT = ___ARG2;
                if( 0 == *input) {
                    int i;
                    input++;
                    for(i=0; i < input_size - 1; i++)
                        *output++=*input++;
                }else{
                    input_size -= 5;
                    input += 5;
                    if ( lzo1x_decompress(
                            input,
                            (lzo_uint) input_size, 
                            output, 
                            &output_size, 
                            NULL) != LZO_E_OK ){
                        ___SCMOBJ ___temp;
                        ___U8VECTORSHRINK(___ARG2,___FIX_0);
                        ___RESULT = ___FIX(___CTOS_HEAP_OVERFLOW_ERR);
                    }
                }
end-of-c-code
vect
result))))
